import string
from enum import auto, Enum
from typing import Callable, Any

from asserts import assert_parsing_succeeds
from parsing.combinators.many import many
from parsing.parser import S
from parsing.strings.char import char
from parsing.strings.char_in import char_in
from parsing.strings.word import word


class Honorific(Enum):
    MR = auto()
    MS = auto()


class Person:
    def __init__(self, honorific: Honorific, name: str, surname: str) -> None:
        self._honorific = honorific
        self._name = name
        self._surname = surname


def const(value: S) -> Callable[[Any], S]:
    return lambda x: value


alphabetic_character = char_in(string.ascii_letters)
alphabetic_word = ''.join * many(alphabetic_character)
mr = const(Honorific.MR) * (word('Mr.') | word('mr.'))
ms = const(Honorific.MS) * (word('Ms.') | word('ms.'))
honorific = (mr | ms) < char(' ')
name = alphabetic_word < char(' ')
surname = alphabetic_word
person = Person * honorific & name & surname


def test_parse_person() -> None:
    assert_parsing_succeeds(person, 'Mr. John revelator')


def test_parse_person_parse_lowercase() -> None:
    assert_parsing_succeeds(person, 'Mr. john Revelator')


def test_parse_surname_parses_lowercase_and_no_space() -> None:
    assert_parsing_succeeds(surname, 'john ').with_remainder(' ')


def test_parse_name_parses_uppercase_and() -> None:
    assert_parsing_succeeds(name, 'John ')


def test_parse_mr() -> None:
    assert_parsing_succeeds(honorific, 'Mr. ').with_result(Honorific.MR)


def test_parse_mr_parses_lowercase() -> None:
    assert_parsing_succeeds(honorific, 'mr. ')


def test_parse_ms() -> None:
    assert_parsing_succeeds(honorific, 'Ms. ').with_result(Honorific.MS)


def test_parse_ms_parses_lowercase() -> None:
    assert_parsing_succeeds(honorific, 'ms. ').with_result(Honorific.MS)



from asserts import assert_parsing_succeeds, assert_parsing_fails
from parsing.combinators.many import many
from parsing.parser import Parser, U, S
from parsing.strings.word import word


def many_till(parser: Parser[U], until: Parser[S]) -> Parser[list[U]]:
    return many(parser) < until


def test_that_many_till_matches_many_and_consumes_delimiter() -> None:
    assert_parsing_succeeds(many_till(word('a'), word('b')), 'aabc').with_result(['a', 'a']).with_remainder('c')


def test_that_many_till_fails_on_nothing() -> None:
    assert_parsing_fails(many_till(word('a'), word('b')), 'b')


def test_that_many_till_fails_when_lookahead_fails() -> None:
    assert_parsing_fails(many_till(word('a'), word('b')), 'a')



from typing import cast

from asserts import assert_parsing_succeeds, assert_parsing_fails
from parsing.parser import Parser, S, U, ParseResults, CouldNotParse
from parsing.strings.word import word


def lookahead(parser: Parser[S], look_for: Parser[U]) -> Parser[S]:
    def parser_(to_parse: str) -> ParseResults[S] | CouldNotParse:
        result_combined = (parser < look_for)(to_parse)
        if isinstance(result_combined, CouldNotParse):
            return result_combined
        return parser(to_parse)

    return Parser(parser_)


def test_that_lookahead_looks_ahead() -> None:
    parser = lookahead(word('a'), look_for=word('b'))

    assert_parsing_succeeds(parser, 'ab').with_result('a')


def test_that_lookahead_fails_when_lookahead_not_ahead() -> None:
    parser = lookahead(word('a'), look_for=word('b'))
    expected_failure_message = cast(CouldNotParse, word('b')('')).reason

    assert_parsing_fails(parser, 'a').with_reason(expected_failure_message)


def test_that_lookahead_keeps_remainder() -> None:
    parser = lookahead(word('a'), look_for=word('b'))

    assert_parsing_succeeds(parser, 'ab').with_remainder('b')



from asserts import assert_parsing_fails, assert_parsing_succeeds
from parsing.check_for_empty_string import check_for_empty_string
from parsing.parser import Parser, ParseResults, CouldNotParse
from parsing.strings.pop_one_character import pop_one_character


def char(c: str) -> Parser[str]:
    @check_for_empty_string
    def parser(to_parse: str) -> ParseResults[str] | CouldNotParse:
        if to_parse[0] != c:
            return CouldNotParse(f'String "{to_parse}" does not start with "{c}"')

        return pop_one_character(to_parse)

    return Parser(parser)


def test_that_empty_strings_do_not_parse() -> None:
    assert_parsing_fails(char('h'), '').with_reason('String to parse is empty')


def test_that_parsing_a_different_character_fails() -> None:
    h_parser = char('h')
    assert_parsing_fails(h_parser, 'peep').with_reason('String "peep" does not start with "h"')


def test_that_parsing_h_succeeds() -> None:
    h_parser = char('h')
    assert_parsing_succeeds(h_parser, 'h').with_result('h')


def test_that_parsing_h_with_remainder_gives_remainder() -> None:
    h_parser = char('h')
    assert_parsing_succeeds(h_parser, 'hoi').with_remainder('oi')



from typing import TypeVarTuple, Callable, overload, TypeVar

from asserts import assert_parsing_succeeds
from parsing.parser import Parser, S, T, ParseResults, CouldNotParse, MappedParser
from parsing.strings.char import char

Ts = TypeVarTuple('Ts')


U = TypeVar('U')


@overload
def new_and(left: MappedParser[S, T], right: Parser[T]) -> Parser[S]:
    pass


@overload
def new_and(left: MappedParser[S, T, U, *Ts], right: Parser[T]) -> MappedParser[S, U, *Ts]:
    pass


def new_and(left, right):
    if left.is_multi_arg:
        def parser_(to_parse: str) -> ParseResults[Callable[[U, *Ts], S]] | CouldNotParse:
            left_result = left(to_parse)
            if isinstance(left_result, CouldNotParse):
                return left_result

            parsed_function = left_result.result
            remainder = left_result.remainder
            right_result = right(remainder)
            if isinstance(right_result, CouldNotParse):
                return right_result

            return ParseResults(
                result=lambda *ts: parsed_function(right_result.result, *ts),
                remainder=right_result.remainder
            )

        return MappedParser(parser_)

    def parser(to_parse: str) -> ParseResults[S] | CouldNotParse:
        left_result = left(to_parse)
        if isinstance(left_result, CouldNotParse):
            return left_result

        parsed_function = left_result.result
        remainder = left_result.remainder
        right_result = right(remainder)
        if isinstance(right_result, CouldNotParse):
            return right_result

        return ParseResults(
            result=parsed_function(right_result.result),
            remainder=right_result.remainder
        )

    return Parser(parser)


def test_new_and_deals_with_callables() -> None:
    def plus(left: str, right: str) -> str:
        return f'{left} plus {right}'

    a = char('a')
    b = char('b')
    parser = plus * a & b
    assert_parsing_succeeds(parser, 'abc').with_result('a plus b').with_remainder('c')


def test_new_and_deals_with_callables_with_more_arguments() -> None:
    def plus(left: str, middle: str, right: str) -> str:
        return f'{left} plus {middle} plus {right}'

    a = char('a')
    b = char('b')
    c = char('c')
    parser = plus * a & b & c
    assert_parsing_succeeds(parser, 'abc').with_result('a plus b plus c')



from asserts import assert_parsing_fails, assert_parsing_succeeds
from parsing.check_for_empty_string import check_for_empty_string
from parsing.parser import CouldNotParse, Parser, ParseResults
from parsing.strings.pop_one_character import pop_one_character


@check_for_empty_string
def _any_char(to_parse: str) -> ParseResults[str] | CouldNotParse:
    return pop_one_character(to_parse)


any_char = Parser(_any_char)


def test_that_any_char_fails_on_empty_string() -> None:
    assert_parsing_fails(any_char, '').with_reason('String to parse is empty')


def test_that_any_char_succeeds_on_a_character() -> None:
    assert_parsing_succeeds(any_char, 'abc').with_result('a').with_remainder('bc')



import string

from asserts import assert_parsing_succeeds
from parsing.fmap import to_int
from parsing.strings.char import char
from parsing.strings.char_in import char_in
from hacky_function_composition import o
from parsing.combinators.many import many


digit = char_in(string.digits)
nonnegative_integer = to_int /o/ ''.join * many(digit)


def negate(x: int) -> int:
    return -x


negative_integer = negate * (char('-') > nonnegative_integer)
integer = nonnegative_integer | negative_integer


def test_that_we_can_parse_single_digits() -> None:
    assert_parsing_succeeds(integer, '3').with_result(3)


def test_that_we_can_parse_more_digits() -> None:
    assert_parsing_succeeds(integer, '33').with_result(33)


def test_that_we_can_parse_negative_integers() -> None:
    assert_parsing_succeeds(integer, '-3').with_result(-3)



import functools
from typing import Callable

from parsing.parser import CouldNotParse, ParseResults, U


def check_for_empty_string(
    parse_function: Callable[[str], ParseResults[U] | CouldNotParse],
) -> Callable[[str], ParseResults[U] | CouldNotParse]:
    @functools.wraps(parse_function)
    def parser_(to_parse: str) -> ParseResults[U] | CouldNotParse:
        if len(to_parse) == 0:
            return CouldNotParse('String to parse is empty')
        return parse_function(to_parse)

    return parser_



from asserts import assert_parsing_succeeds, assert_parsing_fails
from parsing.check_for_empty_string import check_for_empty_string
from parsing.parser import U, Parser, ParseResults, CouldNotParse
from parsing.strings.pop_one_character import pop_one_character
from parsing.strings.char import char


def char_does_not_match(parser: Parser[U]) -> Parser[str]:
    @check_for_empty_string
    def parser_(to_parse: str) -> ParseResults[str] | CouldNotParse:
        result = parser(to_parse)
        if not isinstance(result, CouldNotParse):
            return CouldNotParse()

        return pop_one_character(to_parse)

    return Parser(parser_)


def test_parser_matches_stuff_not_matching_parser() -> None:
    parser = char_does_not_match(char('a'))

    assert_parsing_succeeds(parser, 'b').with_result('b')


def test_parser_does_not_match_parser() -> None:
    parser = char_does_not_match(char('a'))

    assert_parsing_fails(parser, 'a')



import string
from typing import Any

from asserts import assert_parsing_succeeds
from parsing.combinators.many import many
from parsing.combinators.separated_by import some_separated_by
from parsing.integer.integer import integer
from parsing.recursive_parser import RecursiveParser
from parsing.strings.char import char
from parsing.strings.char_in import char_in
from parsing.strings.word import word


valid_json_string_chars = string.ascii_letters + 'idk'
json_string = ''.join * ((char('"') > many(char_in(valid_json_string_chars))) < char('"'))

_json: RecursiveParser[Any] = RecursiveParser()
array = (char('[') > some_separated_by(_json.parser, word(', '))) < char(']')
key_value_pair = (lambda t, s: (t, s)) * (json_string < word(': ')) & _json.parser
json_object = dict * (
    (char('{') > some_separated_by(key_value_pair, word(', '))) < char('}')
)
_json.parser = integer | array | json_string | json_object
json = _json.parser


def test_integer() -> None:
    assert_parsing_succeeds(json, '3').with_result(3)


def test_string() -> None:
    assert_parsing_succeeds(json_string, '"hi"').with_result('hi')


def test_empty_array() -> None:
    assert_parsing_succeeds(json, '[]').with_result([])


def test_array_with_integer() -> None:
    assert_parsing_succeeds(json, '[3]')


def test_array_with_multiple_integers() -> None:
    assert_parsing_succeeds(json, '[3, 4]')


def test_nested_array() -> None:
    assert_parsing_succeeds(json, '[3, []]').with_result([3, []])


def test_empty_object() -> None:
    assert_parsing_succeeds(json, '{}').with_result({})


def test_object_with_one_item() -> None:
    assert_parsing_succeeds(json, '{"a": 3}').with_result({'a': 3})


def test_object_with_two_items() -> None:
    assert_parsing_succeeds(json, '{"a": 3, "b": 4}').with_result({'a': 3, 'b': 4})



from inspect import signature
from typing import Callable, TypeVarTuple, overload, TypeVar, TypeGuard

from asserts import assert_parsing_succeeds, assert_parsing_fails
from parsing.parser import Parser, T, S, ParseResults, CouldNotParse, MappedParser
from parsing.strings.char import char

Ts = TypeVarTuple('Ts')
U = TypeVar('U')


def to_int(string: str) -> int:
    return int(string)


def number_of_arguments(function: Callable) -> int:
    if function in (int, str, float, dict, set, list, tuple, bool):
        return 1
    return len(signature(function).parameters)


def accepts_single_argument(
    function: Callable[[T, U, *Ts], S] | Callable[[T], S],
) -> TypeGuard[Callable[[T], S]]:
    return number_of_arguments(function) == 1


def accept_many_arguments(
    function: Callable[[T, U, *Ts], S] | Callable[[T], S],
) -> TypeGuard[Callable[[T, U, *Ts], S]]:
    return not accepts_single_argument(function)


@overload
def fmap(function: Callable[[T], S], parser: Parser[T]) -> Parser[S]:
    pass


@overload
def fmap(function: Callable[[T, U, *Ts], S], parser: Parser[T]) -> MappedParser[S, U, *Ts]:
    pass


def fmap(
    function,
    parser,
):
    if accepts_single_argument(function):
        def parser_(to_parse: str) -> ParseResults[S] | CouldNotParse:
            result = parser(to_parse)
            if isinstance(result, CouldNotParse):
                return result
            return ParseResults(
                result=function(result.result),
                remainder=result.remainder
            )
        return Parser(parser_)

    if accept_many_arguments(function):
        def parser_1(to_parse: str) -> ParseResults[Callable[[U, *Ts], S]] | CouldNotParse:
            result = parser(to_parse)
            if isinstance(result, CouldNotParse):
                return result
            return ParseResults(
                result=lambda u, *ts: function(result.result, u, *ts),
                remainder=result.remainder
            )

        if number_of_arguments(function) > 2:
            p = MappedParser(parser_1)
            p.is_multi_arg = True
            return p
        else:
            return MappedParser(parser_1)

    raise Exception('Function should accept either one or many arguments')


def test_that_fmap_still_fails_to_parse_unparsable_stuff() -> None:
    parser = to_int * char('3')

    assert_parsing_fails(parser, 'h')


def test_that_fmap_passes_on_failure_reason() -> None:
    parse_to_map_over = char('3')
    parser = to_int * parse_to_map_over
    to_parse = 'h'
    inner_result = parse_to_map_over(to_parse)
    assert isinstance(inner_result, CouldNotParse)
    inner_reason = inner_result.reason

    assert_parsing_fails(parser, to_parse).with_reason(inner_reason)


def test_that_fmap_successfully_parses_parsable_stuff() -> None:
    parser = to_int * char('3')

    assert_parsing_succeeds(parser, '3')


def test_that_fmap_maps_parsed_stuff() -> None:
    parser = to_int * char('3')

    assert_parsing_succeeds(parser, '3').with_result(3)


def test_with_a_different_function() -> None:
    parser = (lambda x: x + 90) * (to_int * char('3'))

    assert_parsing_succeeds(parser, '3').with_result(93)



from asserts import assert_parsing_succeeds
from parsing.strings.char import char
from parsing.parser import Parser, S, T
from parsing.strings.word import word


def ignore_left(left: Parser[T], right: Parser[S]) -> Parser[S]:
    return (lambda t, s: s) * left & right


def test_that_ignore_left_parses_both_and_returns_right() -> None:
    left = char('&')
    right = word('the thing I want to parse')

    assert_parsing_succeeds(left > right, '&the thing I want to parse').with_result('the thing I want to parse')



from collections.abc import Container

from asserts import assert_parsing_fails, assert_parsing_succeeds
from parsing.check_for_empty_string import check_for_empty_string
from parsing.parser import Parser, ParseResults, CouldNotParse
from parsing.strings.pop_one_character import pop_one_character


def char_in(string: Container[str]) -> Parser[str]:
    @check_for_empty_string
    def parser(to_parse: str) -> ParseResults[str] | CouldNotParse:
        if to_parse[0] not in string:
            return CouldNotParse()

        return pop_one_character(to_parse)

    return Parser(parser)


def test_char_in_fails_on_no_characters_specified() -> None:
    assert_parsing_fails(char_in(''), 'whatever')


def test_char_in_succeeds_on_right_character() -> None:
    assert_parsing_succeeds(char_in('w'), 'w').with_result('w')


def test_char_in_keeps_remainder() -> None:
    assert_parsing_succeeds(char_in('w'), 'whatever').with_remainder('hatever')


def test_char_in_only_parses_one() -> None:
    assert_parsing_succeeds(char_in('i'), 'ii').with_result('i')



from asserts import assert_parsing_fails, assert_parsing_succeeds
from parsing.parser import Parser, ParseResults, CouldNotParse


def word(word_to_parse_for: str) -> Parser[str]:
    def parser(to_parse: str) -> ParseResults[str] | CouldNotParse:
        if not to_parse.startswith(word_to_parse_for):
            return CouldNotParse(f'String "{to_parse}" does not start with "{word_to_parse_for}"')
        return ParseResults(word_to_parse_for, to_parse[len(word_to_parse_for):])

    return Parser(parser)


def test_that_parsing_a_different_character_fails() -> None:
    hoi_parser = word('hoi')
    assert_parsing_fails(hoi_parser, 'subaru').with_reason('String "subaru" does not start with "hoi"')


def test_that_parsing_h_succeeds() -> None:
    hoi_parser = word('hoi')
    assert_parsing_succeeds(hoi_parser, 'hoi').with_result('hoi').with_remainder('')


def test_that_parsing_h_with_remainder_gives_remainder() -> None:
    hoi_parser = word('hoi')
    assert_parsing_succeeds(hoi_parser, 'hoi hoi').with_remainder(' hoi')



from __future__ import annotations
from dataclasses import dataclass
from typing import Generic

from asserts import assert_parsing_succeeds, assert_parsing_fails
from parsing.strings.char import char
from parsing.strings.digit import digit
from parsing.parser import Parser, T, ParseResults, CouldNotParse


@dataclass(frozen=True)
class Bracketed:
    content: Bracketed | int


ParsedRecursiveLang = int | Bracketed


class RecursiveParser(Generic[T]):
    def __init__(self) -> None:
        self._parser: Parser[T] | None = None

    @property
    def parser(self) -> Parser[T]:
        return Parser(self._parse_function)

    @parser.setter
    def parser(self, parser: Parser[T]) -> None:
        self._parser = parser

    def _parse_function(self, to_parse: str) -> ParseResults[T] | CouldNotParse:
        if self._parser is None:
            raise NotImplementedError
        return self._parser(to_parse)


open = char('(')
close = char(')')

bracketed: RecursiveParser[Bracketed] = RecursiveParser()

recursive_lang = digit | bracketed.parser
bracketed.parser = Bracketed * ((open > recursive_lang) < close)


def test_digit_is_parsed() -> None:
    assert_parsing_succeeds(recursive_lang, '3').with_result(3)


def test_we_can_do_parens() -> None:
    assert_parsing_succeeds(recursive_lang, '(3)').with_result(Bracketed(3))


def test_we_can_do_two_parens() -> None:
    assert_parsing_succeeds(recursive_lang, '((3))').with_result(Bracketed(Bracketed(3)))


def test_parens_must_close() -> None:
    assert_parsing_fails(recursive_lang, '((3)')


def test_we_can_do_many() -> None:
    assert_parsing_succeeds(recursive_lang, '((((3))))')



from asserts import assert_parsing_fails, assert_parsing_succeeds
from parsing.parser import Parser, T, ParseResults, CouldNotParse
from parsing.strings.char import char


def some(parser: Parser[T]) -> Parser[list[T]]:
    def parser_(to_parse: str) -> ParseResults[list[T]]:
        remainder = to_parse
        result: list[T] = []
        while remainder:
            one_parse = parser(remainder)
            if isinstance(one_parse, CouldNotParse):
                break
            result.append(one_parse.result)
            remainder = one_parse.remainder

        return ParseResults(result=result, remainder=remainder)

    return Parser(parser_)


def many(parser: Parser[T]) -> Parser[list[T]]:
    return (lambda t, ts: [t, *ts]) * parser & some(parser)


def test_some_parses_one() -> None:
    parser = some(char('a'))

    assert_parsing_succeeds(parser, 'a').with_result(['a'])


def test_some_parses_more() -> None:
    parser = some(char('a'))

    assert_parsing_succeeds(parser, 'aa').with_result(['a', 'a'])


def test_some_parses_none() -> None:
    parser = some(char('a'))

    assert_parsing_succeeds(parser, 'h').with_result([]).with_remainder('h')


def test_many_fails_to_parse_unparsable() -> None:
    parser = many(char('a'))

    assert_parsing_fails(parser, 'b')


def test_many_parses_one() -> None:
    parser = many(char('a'))

    assert_parsing_succeeds(parser, 'a').with_result(['a'])


def test_many_parses_two() -> None:
    parser = many(char('a'))

    assert_parsing_succeeds(parser, 'aa').with_result(['a', 'a'])



from asserts import assert_parsing_fails, assert_parsing_succeeds
from parsing.strings.char import char
from parsing.parser import Parser, T, S, ParseResults, CouldNotParse


def or_2(parser_1: Parser[T], parser_2: Parser[S]) -> Parser[T | S]:
    def parser(to_parse: str) -> ParseResults[T | S] | CouldNotParse:
        try_1 = parser_1(to_parse)
        if not isinstance(try_1, CouldNotParse):
            return try_1
        return parser_2(to_parse)

    return Parser(parser)


def test_parsing_neither_fails() -> None:
    a = char('a')
    b = char('b')
    parser = a | b
    assert_parsing_fails(parser, 'c')


def test_parsing_first_succeeds() -> None:
    a = char('a')
    b = char('b')
    parser = a | b
    assert_parsing_succeeds(parser, 'a').with_result('a')


def test_parsing_second_succeeds() -> None:
    a = char('a')
    b = char('b')
    parser = a | b
    assert_parsing_succeeds(parser, 'b').with_result('b')


def test_that_remainder_remains() -> None:
    a = char('a')
    b = char('b')
    parser = a | b
    assert_parsing_succeeds(parser, 'bingo').with_remainder('ingo')



from asserts import assert_parsing_succeeds, assert_parsing_fails
from parsing.combinators.lookahead import lookahead
from parsing.combinators.many import many
from parsing.parser import Parser, U, S
from parsing.strings.word import word


def many_till_exclusive(parser: Parser[U], until: Parser[S]) -> Parser[list[U]]:
    return lookahead(many(parser), look_for=until)


def test_that_many_till_exclusive_matches_many_and_keeps_remainder() -> None:
    assert_parsing_succeeds(many_till_exclusive(word('a'), word('b')), 'aab').with_result(['a', 'a']).with_remainder('b')


def test_that_many_till_exclusive_fails_on_nothing() -> None:
    assert_parsing_fails(many_till_exclusive(word('a'), word('b')), 'b')


def test_that_many_till_exclusive_fails_when_lookahead_fails() -> None:
    assert_parsing_fails(many_till_exclusive(word('a'), word('b')), 'a')



from asserts import assert_parsing_succeeds
from parsing.parser import Parser, S, T
from parsing.strings.word import word


def ignore_right(left: Parser[T], right: Parser[S]) -> Parser[T]:
    return (lambda t, s: t) * left & right


def test_that_ignore_right_parses_both_and_returns_left() -> None:
    left = word('(This I want)')
    right = word('(This I ignore)')

    assert_parsing_succeeds(left < right, '(This I want)(This I ignore)').with_result('(This I want)').with_remainder('')



import string

from asserts import assert_parsing_fails, assert_parsing_succeeds
from parsing.fmap import to_int
from parsing.strings.char_in import char_in


digit = to_int * char_in(string.digits)


def test_that_digit_does_not_parse_letters() -> None:
    assert_parsing_fails(digit, 'h')


def test_that_digit_does_parse_digits() -> None:
    assert_parsing_succeeds(digit, '3').with_result(3).with_remainder('')


def test_that_digit_does_parse_digits_with_remainder() -> None:
    assert_parsing_succeeds(digit, '34').with_result(3).with_remainder('4')


def test_that_digit_parses_all_digits() -> None:
    for d in string.digits:
        assert_parsing_succeeds(digit, d)



from typing import Any

from asserts import assert_parsing_succeeds, assert_parsing_fails
from parsing.combinators.many import some
from parsing.parser import Parser, T
from parsing.strings.char import char
from parsing.strings.word import word


def separated_by(parser: Parser[T], separator: Parser[Any]) -> Parser[list[T]]:
    return (lambda t, ts: [t, *ts]) * parser & some(separator > parser)


nothing = word('')


def some_separated_by(parser: Parser[T], separator: Parser[Any]) -> Parser[list[T]]:
    return separated_by(parser, separator) | ((lambda x: []) * nothing)


def test_separated_by_fails_on_no_match() -> None:
    separator = char(',')
    parser = separated_by(word('hi'), separator)
    assert_parsing_fails(parser, '')


def test_separated_by_succeeds_on_one_match() -> None:
    separator = char(',')
    parser = separated_by(word('hi'), separator)
    assert_parsing_succeeds(parser, 'hi').with_result(['hi'])


def test_separated_by_succeeds_on_two_matches() -> None:
    separator = char(',')
    parser = separated_by(word('hi'), separator)
    assert_parsing_succeeds(parser, 'hi,hi').with_result(['hi', 'hi'])


def test_some_separated_by_succeeds_on_no_match() -> None:
    separator = char(',')
    parser = some_separated_by(word('hi'), separator)
    assert_parsing_succeeds(parser, '').with_result([])


def test_some_separated_by_succeeds_on_one_match() -> None:
    separator = char(',')
    parser = some_separated_by(word('hi'), separator)
    assert_parsing_succeeds(parser, 'hi').with_result(['hi'])


def test_some_separated_by_succeeds_on_two_matches() -> None:
    separator = char(',')
    parser = some_separated_by(word('hi'), separator)
    assert_parsing_succeeds(parser, 'hi,hi').with_result(['hi', 'hi'])



from collections.abc import Container

from asserts import assert_parsing_succeeds, assert_parsing_fails
from parsing.parser import Parser
from parsing.strings.char_does_not_match import char_does_not_match
from parsing.strings.char_in import char_in


def char_not_in(characters: Container[str]) -> Parser[str]:
    return char_does_not_match(char_in(characters))


def test_char_not_in_matches_single_character() -> None:
    parser = char_not_in('abcdefg')
    assert_parsing_succeeds(parser, 'h').with_result('h')


def test_char_not_in_does_not_match_specified_character() -> None:
    parser = char_not_in('abcdefg')
    assert_parsing_fails(parser, 'a')


def test_char_not_in_does_not_match_empty_string() -> None:
    parser = char_not_in('adsf')
    assert_parsing_fails(parser, '')
